library IEEE,mylib;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use mylib.DCOL_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity ddr_w_queue is
    generic (swap_byte : boolean := false);
    port (
-- system clock input
        clk: in STD_LOGIC;
        n_strobe: in STD_LOGIC;
-- reset input, should be connected to the inverted LOCK output of dll
        reset: in STD_LOGIC;
        run: in STD_LOGIC;
				circular : IN std_logic; -- circular mode
				skip : IN std_logic; -- skip sorting
				DCON_en : IN std_logic_vector(12 downto 1);
-- fill memory with random numbers
        mem_test: in STD_LOGIC;
-- sorted channel data
				hit_data : IN std_logic_vector(35 downto 0);
				hit_addr : IN std_logic_vector(5 downto 0);
				hit_rqst : IN std_logic_vector(3 downto 0);
				hit_end : IN std_logic_vector(3 downto 0);
				hit_gnt : OUT std_logic_vector(3 downto 0);
-- event data in
				evt_end : IN std_logic; -- mark the end of a event in non circular mode
				evt_we : IN std_logic;
				evt_data_in : IN std_logic_vector(7 downto 0);
-- DDR access control signals
				ddr_w_we: out std_logic;
        ddr_w_data: out STD_LOGIC_VECTOR (35 downto 0); -- (35 downto 32) low active byte enable for vme write
				ddr_req: out std_logic;
				ddr_read: out std_logic;
				ddr_gnt: in std_logic;
        ddr_length: out STD_LOGIC_VECTOR (3 downto 0);
        ddr_addr: out STD_LOGIC_VECTOR (22 downto 0);
-- sorterB input control signals
				sB_addr : IN std_logic_vector(3 downto 0); -- sorterB data request group address
				sB_rqst : IN std_logic; -- sorterB data request
				sB_gnt: out std_logic;
        sB_g_dav: out STD_LOGIC_VECTOR (15 downto 0); -- sorterB group data available.
-- sorterB input control signals
				sC_addr : IN std_logic_vector(3 downto 0); -- sorterC data request channel address
				sC_rqst : IN std_logic; -- sorterC data request
				sC_gnt: out std_logic;
        sC_ch_dav: out STD_LOGIC_VECTOR (11 downto 0); -- sorterC channel data available
-- sorter data input control
        re_hit: out STD_LOGIC; -- writes hit info to sorterB or sorterC
        we_sB: out STD_LOGIC; -- writes hit info to sorterB or sorterC
        we_sC: out STD_LOGIC; -- writes hit info to sorterB or sorterC
        sorter_wa: out STD_LOGIC_VECTOR (5 downto 0); -- sorter write address
-- download data write
        w_data: in STD_LOGIC_VECTOR (31 downto 0);
        evt_wp: in STD_LOGIC_VECTOR (22 downto 0);
        page_reg: in STD_LOGIC_VECTOR (10 downto 0);
				addr : in STD_LOGIC_VECTOR (22 downto 0);
				we_byte : in STD_LOGIC_VECTOR (3 downto 0);
        rfifo_afull: in STD_LOGIC;
        sel_mod: in STD_LOGIC;
        rd_ram: in STD_LOGIC;
        ram_dav: out STD_LOGIC;
        vme_BLT: in STD_LOGIC;
				evt_size : OUT std_logic_vector(31 downto 0)
    );
end ddr_w_queue;

architecture my_arch of ddr_w_queue is
signal strobe : std_logic := '0';
signal idle: std_logic := '1';
signal idle_dl: std_logic := '1';
signal ch_addr : std_logic_vector(5 downto 0) := (others => '0');
signal ld_ch_addr: std_logic := '0';
signal ld_ram_addr: std_logic := '0';
signal src_sel: std_logic_vector(3 downto 0) := (others => '0');
signal hit_cntr: std_logic_vector(4 downto 0) := (others => '0');
signal fifo_dia: std_logic_vector(8 downto 0) := (others => '0');
signal fifo_aa: std_logic_vector(10 downto 0) := (others => '0');
signal fifo_ab: std_logic_vector(8 downto 0) := (others => '0');
signal we_vme_evt: std_logic := '0';
signal we_vme_evt_dl: std_logic := '0';
signal fifo_dib: std_logic_vector(35 downto 0) := (others => '0');
signal fifo_do: std_logic_vector(35 downto 0) := (others => '0');
signal evt_cntr: std_logic_vector(4 downto 0) := (others => '0');
signal sel_evt_byte : std_logic_vector(1 downto 0) := (others => '0');
signal we_hit: std_logic := '0';signal scan: std_logic_vector(3 downto 0) := (others => '0');
signal rst_cycle: std_logic := '0';
signal rst_cntr: std_logic_vector(4 downto 0) := (others => '0');
signal rp: std_logic_vector(3 downto 0) := (others => '0');
signal group_flag: std_logic_vector(5 downto 0) := (others => '0');
signal event_end: std_logic := '0';
signal g_flag: std_logic_vector(5 downto 0) := (others => '0');
signal e_end: std_logic := '0';
signal hit_b4: std_logic := '0';
signal evt_b4 : std_logic := '0';
signal en_out : std_logic := '0';
signal en_out_dl : std_logic := '0';
signal evt_size_di: std_logic_vector(15 downto 0) := (others => '0');
signal no_evt: std_logic := '0';
signal we_evt_size: std_logic := '0';
signal ram_dav_i : std_logic := '0';
signal ram_dav_dl : std_logic := '0';
signal vme_read : std_logic := '0';
signal dec_vme_wc : std_logic := '0';
signal vme_wc : std_logic_vector(1 downto 0) := (others => '0');
-- address file
signal a_buf_do : std_logic_vector(31 downto 0) := (others => '0');
signal a_buf_dia : std_logic_vector(31 downto 0) := (others => '0');
signal a_buf_dib : std_logic_vector(31 downto 0) := (others => '0');
signal a_buf_aa : std_logic_vector(8 downto 0) := (others => '0');
signal a_buf_ab : std_logic_vector(8 downto 0) := (others => '0');
signal we_a_buf_a : std_logic := '0';
signal we_a_buf_b : std_logic := '0';
signal we_wqueuep: std_logic := '0';
signal we_wqueue: std_logic := '0';
signal re_wqueue: std_logic := '0';
signal wqueue_i: std_logic_vector(7 downto 0) := (others => '0');
signal wqueue_o: std_logic_vector(7 downto 0) := (others => '0');
signal r_length: std_logic_vector(1 downto 0) := (others => '0');
signal rqueue_i: std_logic_vector(6 downto 0) := (others => '0');
signal rqueue_o: std_logic_vector(6 downto 0) := (others => '0');
signal rp_rqueue: std_logic_vector(3 downto 0) := (others => '0');
signal we_rqueuep: std_logic := '0';
signal we_rqueue: std_logic := '0';
signal re_rqueue: std_logic := '0';
signal o_cntr: std_logic_vector(3 downto 0) := (others => '0');
signal oqueue_o: std_logic_vector(6 downto 0) := (others => '0');
signal rp_oqueue: std_logic_vector(3 downto 0) := (others => '0');
signal re_oqueuep: std_logic := '0';
signal re_oqueue: std_logic := '0';
signal o_req : std_logic := '0';
signal vme_r_req : std_logic := '0';
signal mem_rqst : std_logic_vector(1 downto 0) := (others => '0');
signal ddr_read_i : std_logic := '0';
signal w_req : std_logic := '0';
signal r_req : std_logic := '0';
signal sB_gnt_i : std_logic := '0';
signal sC_gnt_i : std_logic := '0';
-- random number memory test signals
signal start_ram_dav : std_logic := '0';
signal ec_lfsr : std_logic := '0';
signal ec_ddr_addr : std_logic := '0';
signal mem_test_dl : std_logic := '0';
signal start_rd_addr : std_logic_vector(2 downto 0) := (others => '0');
signal ram_cntr : std_logic_vector(3 downto 0) := (others => '0');
signal rp_wqueue : std_logic_vector(3 downto 0) := (others => '0');
signal lfsr : std_logic_vector(31 downto 0) := (others => '0');
signal tst_addr : std_logic_vector(31 downto 0) := (others => '0');
begin
ld_ch_addr <= '0';
sB_g_dav <= x"0000";
sC_ch_dav <= x"000";-- when skip = '1' else DCON_en;
ram_dav <= ram_dav_i;
g_g_flag : for i in 0 to 5 generate
  i_g_flag : RAM16X1D
   port map (
      DPO => g_flag(i),     -- Read-only 1-bit data output for DPRA
      A0 => fifo_aa(4),       -- R/W address[0] input bit
      A1 => '0',       -- R/W address[1] input bit
      A2 => '0',       -- R/W address[2] input bit
      A3 => '0',       -- R/W ddress[3] input bit
      D => group_flag(i),         -- Write 1-bit data input
      DPRA0 => rp(2), -- Read-only address[0] input bit
      DPRA1 => '0', -- Read-only address[1] input bit
      DPRA2 => '0', -- Read-only address[2] input bit
      DPRA3 => '0', -- Read-only address[3] input bit
      WCLK => clk,   -- Write clock input
      WE => we_hit        -- Write enable input
   );
end generate;
i_evt_end : RAM16X1D
   port map (
      DPO => e_end,     -- Read-only 1-bit data output for DPRA
      A0 => fifo_ab(2),       -- R/W address[0] input bit
      A1 => '0',       -- R/W address[1] input bit
      A2 => '0',       -- R/W address[2] input bit
      A3 => '0',       -- R/W ddress[3] input bit
      D => event_end,         -- Write 1-bit data input
      DPRA0 => rp(2), -- Read-only address[0] input bit
      DPRA1 => '0', -- Read-only address[1] input bit
      DPRA2 => '0', -- Read-only address[2] input bit
      DPRA3 => '0', -- Read-only address[3] input bit
      WCLK => clk,   -- Write clock input
      WE => we_vme_evt        -- Write enable input
   );
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
		if(strobe = '1')then
			if(reset = '1' or rst_cycle = '0')then
				rst_cntr <= (others => '0');
			else
				rst_cntr <= rst_cntr + 1;
			end if;
      if(reset = '1')then
				rst_cycle <= '1';
			elsif(rst_cntr = "11101")then
				rst_cycle <= '0';
			end if;
			idle_dl <= idle or rst_cycle;
			we_hit <= not idle_dl and not rst_cycle;
			if(we_hit = '0')then
				hit_cntr(3 downto 0) <= x"0";
			else
				hit_cntr(4 downto 0) <= hit_cntr(4 downto 0) + 1;
			end if;
      if(rst_cycle = '1')then
				src_sel <= x"0";
				hit_gnt <= x"0";
			elsif(idle = '1')then
				if(hit_rqst(3) = '1')then
					src_sel <= x"c";
					hit_gnt <= x"8";
				elsif(hit_rqst(2) = '1')then
					src_sel <= "10" & hit_addr(5 downto 4);
					hit_gnt <= x"4";
				elsif(hit_rqst(1) = '1')then
					src_sel <= "01" & hit_addr(3 downto 2);
					hit_gnt <= x"2";
				else
					src_sel <= "00" & hit_addr(1 downto 0); 
					hit_gnt <= "000" & hit_rqst(0);
				end if;
      elsif(hit_cntr(3 downto 0) = x"d")then
				hit_gnt <= x"0";
			end if;
      if(rst_cycle = '1')then
				idle <= '1';
			elsif(run = '0' or hit_cntr(3 downto 0) = x"d")then
				idle <= '1';
			elsif(hit_rqst /= x"0")then
				idle <= '0';
			end if;
			case src_sel(3 downto 2) is
				when "00" => fifo_dia(7 downto 0) <= hit_data(7 downto 0);
								group_flag(5 downto 4) <= hit_data(8) & hit_end(0);
				when "01" => fifo_dia(7 downto 0) <= hit_data(16 downto 9);
								group_flag(5 downto 4) <= hit_data(17) & hit_end(1);
				when "10" => fifo_dia(7 downto 0) <= hit_data(25 downto 18);
								group_flag(5 downto 4) <= hit_data(26) & hit_end(2);
				when others => fifo_dia(7 downto 0) <= hit_data(34 downto 27);
								group_flag(5 downto 4) <= hit_data(35) & hit_end(3);
			end case;
			group_flag(3 downto 0) <= src_sel;
			if(evt_we = '1')then
				event_end <= evt_end;
			end if;
			if(rst_cycle = '1')then
				evt_cntr <= (others => '0');
			elsif(evt_we = '0')then
				evt_cntr(3 downto 0) <= x"0";
			else
				evt_cntr <= evt_cntr + 1;
			end if;
			if(rst_cycle = '1')then
				we_vme_evt <= '0';
			elsif(run = '0')then
				if(mem_test = '1')then
					we_vme_evt <= ec_lfsr;
				elsif(addr(12) = '1' and (we_byte(3) = '1' or (we_byte /= x"0" and vme_BLT = '0')))then
					we_vme_evt <= '1';
				else
					we_vme_evt <= '0';
				end if;
			elsif(evt_cntr(1 downto 0) = "11")then
				we_vme_evt <= '1';
			else
				we_vme_evt <= '0';
			end if;
			if(run = '0')then
				if(mem_test = '1')then
					fifo_dib(31 downto 0) <= lfsr;
				elsif(addr(12) = '1')then
					if(we_byte(0) = '1')then
						fifo_dib(31 downto 24) <= w_data(31 downto 24);
					end if;
					if(we_byte(1) = '1')then
						fifo_dib(23 downto 16) <= w_data(23 downto 16);
					end if;
					if(we_byte(2) = '1')then
						fifo_dib(15 downto 8) <= w_data(15 downto 8);
					end if;
					if(we_byte(3) = '1')then
						fifo_dib(7 downto 0) <= w_data(7 downto 0);
					end if;
				end if;
				if(mem_test = '1')then
					fifo_dib(35 downto 32) <= x"0";
				elsif(sel_mod = '0' or we_vme_evt = '1')then
					fifo_dib(35 downto 32) <= x"f";
				else
					fifo_dib(35 downto 32) <= fifo_dib(35 downto 32) and not we_byte;
				end if;
			else
				case sel_evt_byte is
					when "00" => fifo_dib(31 downto 24) <= evt_data_in;
					when "01" => fifo_dib(23 downto 16) <= evt_data_in;
					when "10" => fifo_dib(15 downto 8) <= evt_data_in;
					when others => fifo_dib(7 downto 0) <= evt_data_in;
				end case;
				fifo_dib(35 downto 32) <= x"0";
			end if;
		end if;
  end if;
end process;
sel_evt_byte <= evt_cntr(1 downto 0) when swap_byte = false else not evt_cntr(1 downto 0);
process(clk)
begin
  if(clk'event and clk = '1')then
		if(strobe = '1')then
			if(rst_cycle = '1')then
				fifo_ab <= (others => '0');
			elsif(evt_cntr(1 downto 0) = "11")then
				fifo_ab(3 downto 0) <= '1' & evt_cntr(4 downto 2);
			else
				fifo_ab(3 downto 0) <= rp;
			end if;
			if(rst_cycle = '1')then
				rp <= x"0"; -- rp is fifo read pointer, but event data write has priority over data output
			elsif(en_out = '0')then
				if(evt_b4 /= evt_cntr(4))then
					rp <= '1' & evt_b4 & "00";
					evt_b4 <= evt_cntr(4);
				else
					rp <= '0' & hit_b4 & "00"; 
					hit_b4 <= hit_cntr(4);
				end if;
			elsif(evt_cntr(1 downto 0) /= "11")then
				rp(1 downto 0) <= rp(1 downto 0) + 1;
			end if;
			if(rst_cycle = '1')then
				en_out <= '0';
			elsif(rp(1 downto 0) = "11" and evt_cntr(1 downto 0) /= "11")then 
				en_out <= '0'; -- output four words a time, if next cycle is event data write, extends output one more cycle 
			elsif(hit_b4 /= hit_cntr(4) or evt_b4 /= evt_cntr(4))then
				en_out <= '1'; -- if there are complete hits or events in the fifo, start output
			end if;
			en_out_dl <= en_out and not rst_cycle;
			if(rst_cycle = '1')then
				ddr_w_we <= '0';
			elsif(run = '0')then
				ddr_w_we <= we_vme_evt;
			else
				ddr_w_we <= en_out_dl and not we_vme_evt;
			end if;
		end if;
  end if;
end process;
fifo_aa <= "000000" & hit_cntr(4 downto 2) & not hit_cntr(1 downto 0);
i_fifo : RAMB16_S9_S36
   generic map(WRITE_MODE_B => "WRITE_FIRST",SIM_COLLISION_CHECK => "NONE")
   port map (
      DOB => fifo_do(31 downto 0),      -- Port B 32-bit Data Output
      DOPB => fifo_do(35 downto 32),    -- Port B 4-bit Parity Output
      ADDRA => fifo_aa,  -- Port A 11-bit Address Input
      ADDRB => fifo_ab,  -- Port B 9-bit Address Input
      CLKA => clk,    -- Port A Clock
      CLKB => clk,    -- Port B Clock
      DIA => fifo_dia(7 downto 0),      -- Port A 8-bit Data Input
      DIB => fifo_dib(31 downto 0),      -- Port B 32-bit Data Input
      DIPA => "0",    -- Port A 1-bit parity Input
      DIPB => fifo_dib(35 downto 32),    -- Port-B 4-bit parity Input
      ENA => strobe,      -- Port A RAM Enable Input
      ENB => strobe,      -- PortB RAM Enable Input
      SSRA => '0',    -- Port A Synchronous Set/Reset Input
      SSRB => '0',    -- Port B Synchronous Set/Reset Input
      WEA => we_hit,      -- Port A Write Enable Input
      WEB => we_vme_evt       -- Port B Write Enable Input
   );
i_a_buf : RAMB16_S36_S36
   generic map(WRITE_MODE_B => "WRITE_FIRST",SIM_COLLISION_CHECK => "NONE")
   port map (
      DOB => a_buf_do,      -- Port B 32-bit Data Output
      ADDRA => a_buf_aa,  -- Port A 9-bit Address Input
      ADDRB => a_buf_ab,  -- Port B 9-bit Address Input
      CLKA => clk,    -- Port A Clock
      CLKB => clk,    -- Port B Clock
      DIA => a_buf_dia,      -- Port A 32-bit Data Input
      DIB => a_buf_dib,      -- Port B 32-bit Data Input
      DIPA => x"0",    -- Port A 4-bit parity Input
      DIPB => x"0",    -- Port-B 4-bit parity Input
      ENA => strobe,      -- Port A RAM Enable Input
      ENB => '1',      -- PortB RAM Enable Input
      SSRA => '0',    -- Port A Synchronous Set/Reset Input
      SSRB => '0',    -- Port B Synchronous Set/Reset Input
      WEA => we_a_buf_a,      -- Port A Write Enable Input
      WEB => we_a_buf_b      -- Port B Write Enable Input
   );
a_buf_aa(8 downto 5) <= x"0";
a_buf_ab(8 downto 6) <= "000";
a_buf_dia(31 downto 23) <= "000000000";
a_buf_dib(31 downto 23) <= "000000000";
a_buf_dib(1 downto 0) <= "00";
process(clk)
begin
  if(clk'event and clk = '1')then
		if(strobe = '1')then
			if(rst_cycle = '1')then
				a_buf_aa(4 downto 0) <= rst_cntr;
			elsif(ld_ch_addr = '1')then
				a_buf_aa(4 downto 0) <= "01100";
			else
				a_buf_aa(4 downto 0) <= "01110";
			end if;
			if(sel_mod = '0')then
				ld_ram_addr <= '1';
			elsif(we_a_buf_a = '1')then
				ld_ram_addr <= '0';
			end if;
			mem_test_dl <= mem_test;
			if(mem_test = '1')then
				we_a_buf_a <= not mem_test_dl;
			else
				we_a_buf_a <= (not reset and rst_cycle) or ld_ch_addr or (ld_ram_addr and sel_mod);
			end if;
			if(rst_cycle = '1' or mem_test = '1')then
				we_wqueuep <= '0';
				if(rst_cntr(3 downto 0) = x"d")then
					a_buf_dia(22 downto 0) <= "010" & x"00000";
				else
					a_buf_dia(22 downto 0) <= "00" & rst_cntr(3 downto 0) & '0' & x"0000";
				end if;
			elsif(run = '0')then
				if(we_wqueuep = '0' and ((addr(12) = '1' and addr(1 downto 0) = "11" and vme_BLT = '1' and we_byte(3) = '1') or 
																(wqueue_i(5 downto 4) /= "00" and vme_BLT = '0')))then
					we_wqueuep <= '1';
				else
					we_wqueuep <= '0';
				end if;
				a_buf_dia(22 downto 0) <= page_reg & addr(11 downto 0);
			else  
				if(ld_ch_addr = '1')then
					a_buf_dia(22 downto 0) <= "00" & ch_addr & "000000000000000";
				elsif(ld_ram_addr = '1')then
					if(circular = '1')then
						a_buf_dia(22 downto 0) <= addr;
					else
						a_buf_dia(22 downto 0) <= page_reg & addr(11 downto 0);
					end if;
				end if;
				if(rp(1 downto 0) = "11" and we_wqueuep = '0')then
					we_wqueuep <= '1';
				else
					we_wqueuep <= '0';
				end if;
			end if;
			we_vme_evt_dl <= we_vme_evt;
			if(run = '0')then
				wqueue_i(7 downto 6) <= "00";
			elsif(rp(3) = '0')then
				wqueue_i(7 downto 6) <= g_flag(5 downto 4);
			else
				wqueue_i(7 downto 6) <= '0' & e_end;
			end if;
			if(run = '1')then
				if(skip = '1' or rp(3) = '1')then
					wqueue_i(3 downto 0) <= x"d";
				else
					wqueue_i(3 downto 0) <= g_flag(3 downto 0);
				end if;
			else
				wqueue_i(3 downto 0) <= x"e";
			end if;
			if(mem_test = '1' or run = '1')then
				wqueue_i(5 downto 4) <= "11";
			elsif(we_wqueuep = '1')then
				wqueue_i(5 downto 4) <= "00";
			elsif(we_vme_evt = '1')then
				wqueue_i(5 downto 4) <= wqueue_i(5 downto 4) + 1;
			elsif(wqueue_i(5 downto 4) /= "00" and vme_BLT = '0')then
				wqueue_i(5 downto 4) <= wqueue_i(5 downto 4) - 1;
			end if;
    end if;
		re_wqueue <= not n_strobe and not rst_cycle and ddr_gnt and not ddr_read_i;
		re_rqueue <= not n_strobe and not rst_cycle and ddr_gnt and a_buf_ab(5);
		we_a_buf_b <= not n_strobe and not rst_cycle and ddr_gnt and (not mem_test or not ddr_read_i);
		if(rst_cycle = '1')then
			ddr_length <= x"f";
		elsif(ddr_read_i = '0')then
			ddr_length <= "00" & wqueue_o(5 downto 4);
		else
			ddr_length <= not a_buf_do(3 downto 0);
		end if;
		if(mem_test = '0')then
			we_wqueue <= we_wqueuep and n_strobe;
		else
			we_wqueue <= tst_addr(1) and tst_addr(0) and we_vme_evt and n_strobe;
		end if;
		if(mem_test = '1')then
			a_buf_dib(22 downto 2) <= a_buf_do(22 downto 2) + 1;
		elsif(ddr_read_i = '1')then
			a_buf_dib(22 downto 17) <= a_buf_do(22 downto 17);
			a_buf_dib(16 downto 4) <= a_buf_do(16 downto 4) + 1;  -- block read can't cross 2048 bytes boarder
			a_buf_dib(3 downto 2) <= "00";
		elsif(wqueue_o(3 downto 0) = x"d")then
			if(wqueue_o(6) = '0')then -- end of event flag
				a_buf_dib(20 downto 2) <= a_buf_do(20 downto 2) + 1;
				if(a_buf_do(20 downto 2) = "1111111111111111111")then
					a_buf_dib(22) <= a_buf_do(22) xor a_buf_do(21);
					a_buf_dib(21) <= a_buf_do(22);
				else
					a_buf_dib(22 downto 21) <= a_buf_do(22 downto 21);
				end if;
			else -- move to next event buffer
				a_buf_dib(20 downto 12) <= a_buf_do(20 downto 12) + 1;
				a_buf_dib(11 downto 2) <= "0000000000";
				if(a_buf_do(20 downto 12) = "111111111")then
					a_buf_dib(22) <= a_buf_do(22) xor a_buf_do(21);
					a_buf_dib(21) <= a_buf_do(22);
				else
					a_buf_dib(22 downto 21) <= a_buf_do(22 downto 21);
				end if;
			end if;
		else
			a_buf_dib(22 downto 17) <= a_buf_do(22 downto 17);
			if(wqueue_o(6) = '0')then -- end of group flag
				a_buf_dib(16 downto 2) <= a_buf_do(16 downto 2) + 1;
			else -- move to next (128 event) group
				a_buf_dib(16 downto 9) <= a_buf_do(16 downto 9) + 1;
				a_buf_dib(8 downto 2) <= "0000000";
			end if;
		end if;
		if(mem_test = '0')then
			ec_ddr_addr <= '1';
		elsif(ddr_gnt = '1')then
			ec_ddr_addr <= ddr_read_i;
		end if;
		if(ec_ddr_addr = '1')then
			ddr_addr(22 downto 4) <= a_buf_do(22 downto 4);
		end if;
		ddr_addr(3 downto 0) <= a_buf_do(3 downto 0);
  end if;
end process;
ddr_w_data <= fifo_do;
i_wqueue : srl_fifo
  generic map (N => 8)
  port map(
	rst => rst_cycle,
	din => wqueue_i,
	clk => clk,
	we => we_wqueue,
	re => re_wqueue,
	rp => rp_wqueue,
	dav => w_req,
	dout => wqueue_o);
g_evt_size: for i in 0 to 1 generate
  i_evt_size : RAMB16_S9_S9
   port map (
      DOB => evt_size(8*i+7 downto 8*i),      -- Port B 8-bit Data Output
      ADDRA => evt_wp(22 downto 12),  -- Port A 11-bit Address Input
      ADDRB => page_reg,  -- Port B 11-bit Address Input
      CLKA => clk,    -- Port A Clock
      CLKB => clk,    -- Port B Clock
      DIA => evt_size_di(8*i+7 downto 8*i),      -- Port A 8-bit Data Input
      DIB => x"00",      -- Port B 8-bit Data Input
      DIPA => "0",    -- Port A 1-bit parity Input
      DIPB => "0",    -- Port-B 1-bit parity Input
      ENA => strobe,      -- Port A RAM Enable Input
      ENB => strobe,      -- PortB RAM Enable Input
      SSRA => '0',    -- Port A Synchronous Set/Reset Input
      SSRB => no_evt,    -- Port B Synchronous Set/Reset Input
      WEA => we_evt_size,      -- Port A Write Enable Input
      WEB => '0'       -- Port B Write Enable Input
   );
end generate;
evt_size(31 downto 16) <= x"0000";
evt_size_di(15) <= '0';
evt_size_di(1 downto 0) <= "00";
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
			if(rst_cycle = '1' or ddr_gnt = '0' or ddr_read_i = '1')then
				we_evt_size <= '0';
			elsif(a_buf_ab(3 downto 0) = x"d" and (wqueue_o(6) = '1' or a_buf_do(11 downto 2) = "111111111111"))then
				evt_size_di(14 downto 4) <= ('0' & a_buf_do(11 downto 2)) + 1;
				we_evt_size <= '1';
			end if;
			if(rst_cycle = '1' or page_reg = evt_wp(22 downto 12))then
				no_evt <= '1';
			else
				no_evt <= '0';
			end if;
			if(rst_cycle = '1' or rp_rqueue(3 downto 2) = "10")then
				sB_gnt_i <= '0';
				sC_gnt_i <= '0';
				we_rqueuep <= '0';
			elsif(sB_rqst = '1' and sB_gnt_i = '0')then
				rqueue_i <= r_length & '0' & sB_addr;
				sB_gnt_i <= '1';
				sC_gnt_i <= '0';
				we_rqueuep <= '1';
			elsif(sC_rqst = '1' and sC_gnt_i = '0')then
				rqueue_i <= r_length & '1' & sC_addr;
				sB_gnt_i <= '0';
				sC_gnt_i <= '1';
				we_rqueuep <= '1';
			else
				sB_gnt_i <= '0';
				sC_gnt_i <= '0';
				we_rqueuep <= '0';
			end if;
			start_ram_dav <= ddr_gnt and ddr_read_i;
			if(ram_dav_i = '0')then
				if(mem_test = '1')then
					ram_cntr <= x"0";
				else
--		    ram_cntr <= '0' & addr(2 downto 0);
					ram_cntr <= '0' & start_rd_addr;
				end if;
			else
				ram_cntr <= ram_cntr + 1;
			end if;
			if(ram_dav_i = '1')then
				start_rd_addr <= "000";
			elsif(ld_ram_addr = '1')then
				start_rd_addr <= addr(2 downto 0);
			end if;
			if(mem_test = '1')then
				if(start_ram_dav = '1')then
					ram_dav_i <= '1';
				elsif(ram_cntr = x"f")then
					ram_dav_i <= '0';
				end if;
			elsif(rd_ram = '0' or (ram_dav_i = '1' and ram_cntr(2 downto 0) = "111"))then
				ram_dav_i <= '0';
			elsif(rfifo_afull = '0' and vme_wc /= "00" and ram_dav_dl = '0')then
				ram_dav_i <= '1';
			end if;
			ram_dav_dl <= ram_dav_i;
			if(rd_ram = '0')then
				vme_wc <= (others => '0');
				dec_vme_wc <= '0';
			else
				if((vme_read = '1' and ddr_gnt = '1') or dec_vme_wc = '1')then
					if(vme_read = '1' and ddr_gnt = '1')then
						if(a_buf_do(3) = '1')then
							vme_wc <= "01";
						else
							vme_wc(1) <= vme_wc(0) or not dec_vme_wc;
							vme_wc(0) <= vme_wc(0) xor dec_vme_wc;
						end if;
					else
						vme_wc(1) <= vme_wc(1) and vme_wc(0);
						vme_wc(0) <= not vme_wc(0);
					end if;
				end if;
				if(ram_dav_i = '1' and ram_dav_dl = '0' and vme_BLT = '1')then
					dec_vme_wc <= '1';
				else
					dec_vme_wc <= '0';
				end if;
			end if;
			if(mem_test = '1' and ddr_gnt = '1' and a_buf_do(3 downto 2) = "11")then
				vme_r_req <= '1';
			elsif(rd_ram = '0' or vme_wc(1) = '1' or (vme_wc(0) = '1' and vme_BLT = '0') or
							 (ddr_gnt = '1' and vme_read = '1'))then
				vme_r_req <= '0';
			else
				vme_r_req <= '1';
			end if;
			if(rst_cycle = '1')then
				ddr_read_i <= '1';
				mem_rqst(0) <= '0';
				a_buf_ab(5 downto 0) <= "000000";
			elsif(ddr_gnt = '1')then
				mem_rqst(0) <= '0';
			elsif(mem_rqst(0) = '0')then
				if(vme_r_req = '1')then
					mem_rqst(0) <= '1';
					ddr_read_i <= '1';
					vme_read <= '1';
					a_buf_ab(5 downto 0) <= "001110";
				elsif(r_req = '1')then
					mem_rqst(0) <= '1';
					ddr_read_i <= '1';
					vme_read <= '0';
					a_buf_ab(5 downto 0) <= '1' & rqueue_o(4 downto 0);
				elsif(w_req = '1')then
					mem_rqst(0) <= '1';
					ddr_read_i <= '0';
					vme_read <= '0';
					a_buf_ab(5 downto 0) <= "00" & wqueue_o(3 downto 0);
				end if;
			end if;
		end if;
		we_rqueue <= we_rqueuep and n_strobe and not rst_cycle;
    mem_rqst(1) <= mem_rqst(0) and not ddr_gnt and not rst_cycle;
    ddr_req <= mem_rqst(1) and not ddr_gnt and not rst_cycle;
  end if;
end process;
sB_gnt <= sB_gnt_i;
sC_gnt <= sC_gnt_i;
ddr_read <= ddr_read_i;
i_rqueue : srl_fifo
  generic map (N => 7)
  port map(
	rst => rst_cycle,
	din => rqueue_i,
	clk => clk,
	we => we_rqueue,
	re => re_rqueue,
	dav => r_req,
	rp => rp_rqueue,
	dout => rqueue_o);
i_oqueue : srl_fifo
  generic map (N => 7)
  port map(
	rst => rst_cycle,
	din => rqueue_o,
	clk => clk,
	we => re_rqueue,
	re => re_oqueue,
	dav => o_req,
	rp => rp_oqueue,
	dout => oqueue_o);
re_oqueue <= strobe and not ram_dav_i and re_oqueuep;
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
			if(rst_cycle = '1' or o_req = '0' or (re_oqueuep = '1' and ram_dav_i = '0'))then
				o_cntr <= x"0";
				we_sB <= '0';
				we_sC <= '0';
			elsif(ram_dav_i = '0')then
				o_cntr <= o_cntr + 1;
				we_sB <= not oqueue_o(4);
				we_sC <= oqueue_o(4);
			end if;
			if(rst_cycle = '0' and oqueue_o(6 downto 5) = o_cntr(3 downto 2) and o_cntr(1 downto 0) = "10")then
				re_oqueuep <= '1';
			elsif(ram_dav_i = '0')then
				re_oqueuep <= '0';
			end if;
			sorter_wa <= oqueue_o(3 downto 0) & o_cntr(1 downto 0);
			re_hit <= o_req and not rst_cycle;
		end if;
  end if;
end process;
-- random number memory test added 8/3/2005
-- random number generator
process(clk)
begin
  if(clk'event and clk = '1')then
		if(strobe = '1')then
			if(mem_test = '1' and rp_wqueue(3 downto 2) /= "10")then
				ec_lfsr <= '1';
			else
				ec_lfsr <= '0';
			end if;
      if(mem_test = '0')then
				tst_addr <= (others => '0');
				lfsr <= (others => '0');
			elsif(ec_lfsr = '1')then
				tst_addr <= tst_addr + 1;
				lfsr <= lfsr(30 downto 0) & not(lfsr(31) xor lfsr(21) xor lfsr(1) xor lfsr(0));
			end if;
		end if;
  end if;
end process;
end my_arch;
