-- due to tRAS minimum 42ns, active to prechage for read cycle must
-- have at least six clock cycles(120MHz). This means a minimum of
-- eight words burst read. This also guarantees active to active is
-- eight clock cycles minimum 
library IEEE,mylib;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.all;
use IEEE.std_logic_unsigned.all;
use mylib.DCOL_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity sdram is
	 generic (sim : boolean := false);
    port (
-- reset input, should be connected to the inverted LOCK output of dll
        reset: in STD_LOGIC;
-- system clock input
        mclk0: in STD_LOGIC;
        mclk90: in STD_LOGIC;
        mclk180: in STD_LOGIC;
        mclk270: in STD_LOGIC;
        n_strobe: in STD_LOGIC;
-- DLL lock input
        locked: in STD_LOGIC;
        mem_test: in STD_LOGIC;
		  mem_status : out STD_LOGIC_VECTOR (15 downto 0);
-- VME read signals
        run: in STD_LOGIC;
		  addr : in STD_LOGIC_VECTOR (22 downto 0);
		  rfifo_data : out STD_LOGIC_VECTOR (31 downto 0);
        rd_ram: in STD_LOGIC;
        ram_dav: in STD_LOGIC;
        re_hit: in STD_LOGIC;
-- memory page register page size = 16KBytes
-- write queue signals
				ddr_gnt: out std_logic;
				ddr_w_we: in std_logic;
				ddr_read: in std_logic;
				ddr_rqst: in std_logic;
        ddr_length: in STD_LOGIC_VECTOR (3 downto 0);
        ddr_addr: in STD_LOGIC_VECTOR (22 downto 0);
        ddr_w_data: in STD_LOGIC_VECTOR (35 downto 0);
-- ram pads
		  mem_pads : INOUT mem_pad
    );
end sdram;

architecture sdram_arch of sdram is
component refresh
	 generic (sim : boolean := false);
    Port ( clk : in std_logic;
           lock : in std_logic;
           init_done : in std_logic;
           start_init : out std_logic;
           ddr_csn : out std_logic;
           ddr_clke : out std_logic;
           ref : out std_logic);
end component;
component ddr_io
    port (
        clk: in STD_LOGIC;
        clkn: in STD_LOGIC;
        pad: inout std_logic;
        t: in STD_LOGIC;
        d0: in STD_LOGIC;
        d1: in STD_LOGIC;
        iclk: in STD_LOGIC;
        iclkn: in STD_LOGIC;
        q0: out std_logic; -- rising edge       
        q1: out std_logic  -- falling edge    
    );
end component;
-- sdram i/o
signal strobe : std_logic := '0';
signal ma: std_logic_vector(14 downto 0) := (others => '0');
signal md: std_logic_vector(31 downto 0) := (others => '0');
signal byte_n: std_logic_vector(3 downto 0) := (others => '0');
signal mq: std_logic_vector(31 downto 0) := (others => '0');
signal mq_q: std_logic_vector(31 downto 0) := (others => '0');
signal mq_q2: std_logic_vector(31 downto 0) := (others => '0');
signal rasn: std_logic := '1';
signal casn: std_logic := '1';
signal wen: std_logic := '1';
signal set_dm: std_logic := '0';
signal casn_dl: std_logic := '1';
signal active_dl: std_logic_vector(3 downto 0) := (others => '0');
signal t_md: std_logic := '1';
-- refresh signal generation
signal ref: std_logic := '0';
signal start_init: std_logic := '0';
-- sdram state machine
signal en_dll: std_logic := '0';
signal rst_dll: std_logic := '0';
signal w_cycle: std_logic := '0';
signal r_cycle: std_logic := '0';
signal vme_read: std_logic := '0';
signal flush_vme_buf: std_logic := '0';
signal wr_burst: std_logic := '0';
signal rd_burst: std_logic := '0';
signal stop_w: std_logic := '0';
signal stop_r: std_logic := '0';
signal we_rfifo: std_logic := '0';
signal active: std_logic := '0';
signal inh_ras: std_logic := '0';
signal ref_rqst: std_logic := '0';
signal init_done: std_logic_vector(1 downto 0) := (others => '1');
signal pc_dl: std_logic := '0';
signal precharge: std_logic := '0';
signal prechage_dl: std_logic := '0';
signal rfc_cntr: std_logic_vector(3 downto 0) := (others => '0');
-- random memory test signals
signal lfsr : std_logic_vector(31 downto 0) := (others => '0');
signal ec_lfsr: std_logic := '0';
signal not_locked: std_logic := '0';
signal rst_rclk: std_logic := '0';
signal rclk0_dcm: std_logic := '0';
signal rclk90_dcm: std_logic := '0';
signal rclk0: std_logic := '0';
signal rclk90: std_logic := '0';
signal rclk180: std_logic := '0';
signal rclk270: std_logic := '0';
signal wfifo_do: std_logic_vector(35 downto 0) := (others => '0');
signal wfifo_wa : std_logic_vector(8 downto 0) := (others => '0');
signal wfifo_ra : std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_wa : std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_ra : std_logic_vector(8 downto 0) := (others => '0');
signal rfifo_vme_wa : std_logic_vector(7 downto 0) := (others => '0');
signal rfifo_vme_ra : std_logic_vector(7 downto 0) := (others => '0');
signal rfifo_hit_wa : std_logic_vector(7 downto 0) := (others => '0');
signal rfifo_hit_ra : std_logic_vector(7 downto 0) := (others => '0');
signal rfifo_data_i: std_logic_vector(31 downto 0) := (others => '0');
signal burst_cnt : std_logic_vector(4 downto 0) := (others => '0');
signal wr_burst_dl: std_logic := '0';
signal ec_rfifo_wa: std_logic := '0';
signal ec_dqs: std_logic := '0';
signal chk_ps: std_logic := '0';
signal t_dqs: std_logic := '0';
signal dqs_in : std_logic_vector(3 downto 0) := (others => '0');
signal dqs_in_qp : std_logic_vector(1 downto 0) := (others => '0');
signal dqs_in_q : std_logic_vector(1 downto 0) := (others => '0');
signal dqs_status : std_logic_vector(3 downto 0) := (others => '0');
signal rclk_tune: std_logic := '0';
signal psincdec: std_logic := '0';
signal psen: std_logic := '0';
signal psdone: std_logic := '0';
signal ps_rdy: std_logic := '0';
signal ps_cntr : std_logic_vector(7 downto 0) := (others => '0');
signal ps_delay : std_logic_vector(7 downto 0) := (others => '0');
signal dqs_out: std_logic := '0';
signal mem_rdy: std_logic := '0';
signal mem_rdyp: std_logic := '0';
signal ps_cntr_q : std_logic_vector(7 downto 0) := (others => '0');
signal ps_rdy_q: std_logic := '0';
signal mem_rdy_q: std_logic := '0';
--attribute LOC : string;
--attribute LOC of i_dqs_in_qp0, i_dqs_in_qp1 : label is "SLICE_X0Y144";
begin
rfifo_data <= rfifo_data_i;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    strobe <= n_strobe;
    if(strobe = '1')then
      mem_status <= mem_rdy_q & init_done(1) & ps_rdy_q & '0' & x"0" & ps_cntr_q;
		end if;
    if(locked = '0')then
      ps_cntr_q <= x"80";
    elsif(psen = '1')then
      ps_cntr_q <= ps_cntr;
		end if;
    mem_rdy_q <= mem_rdy;
    ps_rdy_q <= ps_rdy;
		if(mem_rdy_q = '1' or active_dl(1) = '1')then
			rclk_tune <= '0';
		elsif(init_done(1) = '1' and ref = '1')then
			rclk_tune <= '1';
		end if;
  end if;
end process;
i_wfifo : RAMB16_S36_S36
   generic map(SIM_COLLISION_CHECK => "NONE")
   port map (
      DIA => ddr_w_data(31 downto 0),      -- Port A 32-bit Data Input
      DIPA => ddr_w_data(35 downto 32),    -- Port-A 4-bit parity Input
      ADDRA => wfifo_wa,  -- Port A 10-bit Address Input
      CLKA => mclk0,    -- Port A Clock
      SSRA => '0',    -- Port A Synchronous Set/Reset Input
      ENA => strobe,      -- Port A RAM Enable Input
      WEA => ddr_w_we,      -- Port A Write Enable Input
      DIB => x"00000000",      -- Port B 16-bit Data Input
      DIPB => x"0",    -- Port B 2-bit parity Input
      ADDRB => wfifo_ra,  -- Port B 10-bit Address Input
      CLKB => mclk0,    -- Port B Clock
      SSRB => '0',    -- Port B Synchronous Set/Reset Input
      ENB => '1',      -- PortB RAM Enable Input
      WEB => '0',    -- Port B Write Enable Input
      DOB => wfifo_do(31 downto 0),      -- Port B 32-bit Data Output
      DOPB => wfifo_do(35 downto 32)      -- Port B 4-bit Data Output
   );
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
-- wfifo address counters
		if(strobe = '1')then
			if(reset = '1')then
				wfifo_wa <= (others => '0');
			elsif(ddr_w_we = '1')then
				wfifo_wa <= wfifo_wa + 1;
			end if;	
		end if;
		if(reset = '1')then
			wfifo_ra <= (others => '0');
		elsif(wr_burst = '1')then
			wfifo_ra <= wfifo_ra + 1;
		end if;	
		mq_q(15 downto 0) <= mq(15 downto 0);
		mq_q2 <= mq_q;
		md <= wfifo_do(31 downto 0);
		byte_n <= wfifo_do(35 downto 32);
    if(reset = '1')then
      rfifo_hit_wa <= (others => '0');
    elsif(ec_rfifo_wa = '1' and vme_read = '0')then
      rfifo_hit_wa <= rfifo_hit_wa + 1;
    end if;
    if(flush_vme_buf = '1')then
      rfifo_vme_wa <= (others => '0');
    elsif(ec_rfifo_wa = '1' and vme_read = '1')then
      rfifo_vme_wa <= rfifo_vme_wa + 1;
    end if;
		if(vme_read = '1')then
			rfifo_wa <= '1' & rfifo_vme_wa;
		else
			rfifo_wa <= '0' & rfifo_hit_wa;
		end if;
		we_rfifo <= ec_rfifo_wa;
  end if;
end process;
process(mclk180)
begin
  if(mclk180'event and mclk180 = '1')then
		mq_q(31 downto 16) <= mq(31 downto 16);
  end if;
end process;
i_rfifo: ramb16_s36_s36
     port map(
         dia => x"00000000", 
         dipa => x"0", 
         addra => rfifo_ra,
         clka => mclk0, 
         ssra => '0', 
         ena => strobe, 
         wea => '0', 
         doa => rfifo_data_i, 
         dib =>  mq_q2,
         dipb => x"0",
         addrb => rfifo_wa,
         clkb => mclk0,
         ssrb => '0',
         enb => '1',
         web => we_rfifo);
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
		if(rd_ram = '0' and mem_test = '0')then
			flush_vme_buf <= '1';
    elsif(vme_read = '0')then
			flush_vme_buf <= '0';
    end if;
		if(active_dl(3) = '1')then
			vme_read <= (ma(14) or ma(13) or not run) and r_cycle;
		elsif(ec_rfifo_wa = '0' and we_rfifo = '1')then
			vme_read <= '0';
		end if;
    if(strobe = '1')then
			if(reset = '1')then
				rfifo_hit_ra <= (others => '0');
			elsif(ram_dav = '0' and re_hit = '1')then
				rfifo_hit_ra <= rfifo_hit_ra + 1;
			end if;
			if(flush_vme_buf = '1')then
				rfifo_vme_ra <= (others => '0');
			elsif(ram_dav = '1')then
				rfifo_vme_ra <= rfifo_vme_ra + 1;
			end if;
    end if;
  end if;
end process;
rfifo_ra <= '1' & rfifo_vme_ra when ram_dav = '1' else '0' & rfifo_hit_ra;
i_ec_rfifo_wa: delayN
  generic map(N=> 5) port map(d => rd_burst, ce => '1', c => mclk0, q => ec_rfifo_wa);
process(mclk180,locked)
begin
  if(locked = '0')then
		ec_dqs <= '0';
		chk_ps <= '0';
--    mem_rdy <= '0';
    ps_rdy <= '1';
    ps_cntr <= x"80";
		dqs_status <= x"0";
		psen <= '0';
		if(sim)then
			ps_delay <= x"7e";
		else
			ps_delay <= x"00";
		end if;
  elsif(mclk180'event and mclk180 = '1')then
		ec_dqs <= ec_rfifo_wa;
		chk_ps <= ref;
    ps_rdy <= not psen and (ps_rdy or psdone);
		if(psen = '1')then
			if(psincdec = '1')then
				ps_cntr <= ps_cntr + 1;
			else
				ps_cntr <= ps_cntr - 1;
			end if;
		end if;
		if(psdone = '1' or chk_ps = '1')then
			dqs_status <= x"0";
		elsif(ec_dqs = '1')then
			dqs_status(0) <= dqs_status(0) or not dqs_in_q(0);
			dqs_status(1) <= dqs_status(1) or dqs_in_q(0);
			dqs_status(2) <= dqs_status(2) or not dqs_in_q(1);
			dqs_status(3) <= dqs_status(3) or dqs_in_q(1);
		end if;
		if(chk_ps = '1' and ps_rdy = '1')then
			case dqs_status is
--			when "0000" => null;	 
				when "1001" =>
					psen <= '1';
					psincdec <= '1';
				when "0110" =>
					psen <= '1';
					psincdec <= '0';
				when others => null;	 
			end case;
		else
			psen <= '0';
		end if;
		if(ps_delay(7) = '0' and ref = '1')then
			ps_delay <= ps_delay + 1;
		end if;
  end if;
end process;
mem_rdy <= ps_delay(7);
i_dqs_in_qp0 : FD
   port map (
      Q => dqs_in_qp(0),      -- Data output
      C => rclk270,      -- Clock input
      D => dqs_in(0)       -- Data input
   );
i_dqs_in_qp1 : FD
   port map (
      Q => dqs_in_qp(1),      -- Data output
      C => rclk270,      -- Clock input
      D => dqs_in(1)       -- Data input
   );
process(rclk270)
begin
  if(rclk270'event and rclk270 = '1')then
--    dqs_in_qp <= dqs_in(1 downto 0);
  end if;
end process;
process(rclk180)
begin
  if(rclk180'event and rclk180 = '1')then
    dqs_in_q <= dqs_in_qp;
  end if;
end process;
-- memory address
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
-- this ensures active, read/write and precharge are to the same bank
    if(init_done(1) = '0')then
			ma(14 downto 13) <= '0' & en_dll; 
    elsif(inh_ras = '0')then
      ma(14 downto 13) <= ddr_addr(22 downto 21);
		end if;
    if(init_done(1) = '0')then
-- latency = 2 burst = 8
			ma(12 downto 0) <= "00" & precharge & '0' & rst_dll & "00" & rst_dll & "000" & (rst_dll or en_dll) & rst_dll; 
		elsif(inh_ras = '0')then
      ma(12 downto 0) <= ddr_addr(20 downto 8);
		elsif(active_dl(0) = '1')then
			ma(12 downto 0) <= "00" & precharge & '0' & ddr_addr(7 downto 0) & '0';
		elsif(rd_burst = '1')then 	
			ma(4 downto 1) <= ma(4 downto 1) + 1;
		end if;
  end if;
end process;
-- generate read and write burst signals -------------------------------------------------------
-- burst_cntr
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    if(reset = '1')then
			burst_cnt <= (others => '0');
		elsif(active = '1')then	 
			burst_cnt <= '1' & not ddr_length;
		elsif(burst_cnt(4) = '1')then
      burst_cnt <= burst_cnt + 1;	 
    end if;
		active_dl <= active_dl(2 downto 0) & active;
		rd_burst <= burst_cnt(4) and r_cycle;
		wr_burst <= burst_cnt(4) and w_cycle;
		wr_burst_dl <= wr_burst;
		if(strobe = '1')then
			ddr_gnt <= active_dl(3) or active_dl(2);
		end if;
  end if;
end process;
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    if(reset = '1')then
      t_md <= '1';
      t_dqs <= '1';
			dqs_out <= '0';
			set_dm <= '1';
		else	
      t_md <= not(not casn and w_cycle) and (t_md or t_dqs);
      t_dqs <= not(not casn and w_cycle) and (t_dqs or not casn_dl);
			dqs_out <= not t_dqs;
			set_dm <= not wr_burst_dl;
		end if;	
  end if;
end process;
i_casn_dl: delayN
  generic map(N=> 4) port map(d => casn, ce => '1', c => mclk0, q => casn_dl);
-- sdram state machine -----------------------------------------------------------------------------------------
-- sdram power on initialization states: precharge all banks, enable DLL, reset DLL, precharge all banks,
-- two refresh cycle and set mode register.
i_en_dll: delayN
  generic map(N=> 4) port map(d => start_init, ce => '1', c => mclk0, q => en_dll);
i_rst_dll: delayN
  generic map(N=> 6) port map(d => start_init, ce => '1', c => mclk0, q => rst_dll);
process(mclk0,locked,rasn,casn,wen)
variable state: std_logic_vector(2 downto 0);
begin
state := rasn & casn & wen;
  if(locked = '0')then
    w_cycle <= '0';
    r_cycle <= '0';
    pc_dl <= '0';
    precharge <= '0';
    prechage_dl <= '0';
    rfc_cntr <= "0000";
		inh_ras <= '0';
    active <= '0';
    stop_w <= '0';
    stop_r <= '0';
    rasn <= '1';
    casn <= '1';
    wen <= '1';
    init_done <= "00";
    ref_rqst <= '0';
  elsif(mclk0'event and mclk0 = '1')then
		pc_dl <= start_init or rst_dll or stop_r or stop_w;
    precharge <= pc_dl;
    prechage_dl <= precharge;
-- auto refresh tRFC counter
    if(state = "001")then
			rfc_cntr <= x"8";
		elsif(rfc_cntr(3) = '1')then
			rfc_cntr <= rfc_cntr + 1;
		end if;
-- inhibit RAS after refresh or read/write until tRFC or tPR
    if(rasn = '0' and wen = '1')then
			inh_ras <= '1';
    elsif(prechage_dl = '1' or rfc_cntr(2 downto 1) = "11")then
			inh_ras <= '0';
		end if;
-- latch refresh request
    if(ref = '1')then
      ref_rqst <= '1';
    end if;
-- default to go to NOP
    rasn <= '1';
    casn <= '1';
    wen <= '1';
		active <= '0';
    stop_w <= wr_burst and not burst_cnt(4);
		if(casn = '0' and rd_burst = '1' and (burst_cnt(3 downto 2) = "11" or burst_cnt(4) = '0'))then
			stop_r <= '1';
		else
			stop_r <= '0';
		end if;
    if(active = '1')then
			r_cycle <= ddr_read;
			w_cycle <= not ddr_read;
		elsif(rasn = '0')then
			r_cycle <= '0';
			w_cycle <= '0';
		end if;
    case state is
-- NOP
      when "111" =>
---- goto precharge
        if(precharge = '1')then
          rasn <= '0';
          wen <= '0';
        end if;
---- goto auto refresh
        if(ref_rqst = '1' and inh_ras = '0')then
          rasn <= '0';
          casn <= '0';
        end if;
---- goto set mode
        if(en_dll = '1' or rst_dll = '1')then
          rasn <= '0';
          casn <= '0';
          wen <= '0';
        end if;
---- goto active
        if((ddr_rqst = '1' or rclk_tune = '1') and ref_rqst = '0' and inh_ras = '0')then
          rasn <= '0';
					active <= '1';
        end if;
---- goto write/read
        if(active_dl(0) = '1' or (burst_cnt(4) = '1' and ma(2 downto 1) = "11"))then
          casn <= '0';
          wen <= r_cycle;
        end if;
-- AUTO REFRESH
      when "001" =>
				ref_rqst <= '0';
        init_done <= init_done(0) & '1';
      when others =>
      null;
    end case;
  end if;
end process;
----------------------------------------------------------------------------------------
i_rst_rclk : SRL16
   generic map (
      INIT => X"ffff")
   port map (
      Q => rst_rclk,       -- SRL data output
      A0 => '1',     -- Select[0] input
      A1 => '1',     -- Select[1] input
      A2 => '1',     -- Select[2] input
      A3 => '1',     -- Select[3] input
      CLK => mclk0,   -- Clock input
      D => not_locked        -- SRL data input
   );
not_locked <= not locked;
DCMrclk: dcm 
   generic map(CLKOUT_PHASE_SHIFT => "VARIABLE",
					PHASE_SHIFT => 0)  
	port map	
         ( CLKIN   => mclk90 ,
           CLKFB   => rclk0 ,
           RST     => rst_rclk,
           DSSEN   => '0' ,
           PSINCDEC=> psincdec ,
           PSEN    => psen ,
           PSCLK   => mclk180 ,
			  clk0    => rclk0_dcm,
           CLK90   => rclk90_dcm,
			  PSDONE  => psdone);
rclk180 <= not rclk0;
rclk270 <= not rclk90;
rclk0bufg : BUFG port map (O=>rclk0, I=>rclk0_dcm);
rclk90bufg : BUFG port map (O=>rclk90, I=>rclk90_dcm);
----------------------------------------------------------------------------------------
-- ram pads
process(mclk0)
begin
  if(mclk0'event and mclk0 = '1')then
    mem_pads.we_n <= wen;
    mem_pads.cas_n <= casn;
    mem_pads.ras_n <= rasn;
    mem_pads.ma <= ma(12 downto 0);
		mem_pads.mba <= ma(14 downto 13);
  end if;
end process;
g_dm: for i in 0 to 1 generate
  i_dm: fddrrse
    port map(
          c0 => mclk0,
          c1 => mclk180,
          q => mem_pads.dm(i),
					ce => '1',
          d0 => byte_n(i),
          d1 => byte_n(2+i),
          r => '0',
          s => set_dm);
end generate;
g_mdq: for i in 0 to 15 generate
  ddr_data: ddr_io
    port map(
          clk => mclk0,
          clkn => mclk180,
          pad => mem_pads.mdq(i),
					t => t_md,
          d0 => md(i+16),
          d1 => md(i),
          iclk => rclk0,
          iclkn => rclk180,
          q0 => mq(i),
          q1 => mq(16+i));
end generate g_mdq;
g_dqs: for i in 0 to 1 generate
  i_dqs: ddr_io
    port map(
          clk => mclk270,
          clkn => mclk90,
          pad => mem_pads.dqs(i),
					t => t_md,
          d0 => dqs_out,
          d1 => '0',
          iclk => rclk270,
          iclkn => rclk90,
          q0 => dqs_in(i*2),
          q1 => dqs_in(i*2+1));
end generate;
i_mclk:  fddrrse
    port map(
          d0 => '1',
          d1 => '0',
          c0 => mclk270,
          c1 => mclk90,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mem_pads.mclk);
i_mclk_n:  fddrrse
    port map(
          d0 => '0',
          d1 => '1',
          c0 => mclk270,
          c1 => mclk90,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => mem_pads.mclk_n);
-- instantiate refresh generator
I_refresh: refresh
	generic map(sim => sim)
	PORT MAP(
		clk => mclk0,
		lock => locked,
		init_done => init_done(1),
		start_init => start_init,
		ddr_csn => mem_pads.cs_n,
		ddr_clke => mem_pads.mclke,
		ref => ref
	);
end sdram_arch;
----------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

entity refresh is
	 generic (sim : boolean := false);
    Port ( clk : in std_logic;
           lock : in std_logic;
           init_done : in std_logic;
           start_init : out std_logic;
           ddr_csn : out std_logic;
           ddr_clke : out std_logic;
           ref : out std_logic);
end refresh;

architecture refresh_arch of refresh is
signal clke: std_logic := '0';
signal tc: std_logic := '0';
signal startup: std_logic_vector(4 downto 0) := (others => '0');
signal div: std_logic_vector(9 downto 0) := (others => '0');
begin
ddr_clke <= clke;
-- refresh signal generator
process(clk,lock)
begin
  if(lock = '0')then
    tc <= '0';
    ref <= '0';
    clke <= '0';
    start_init <= '0';
    ddr_csn <= '1';
		if(sim)then
			startup <= "11011";
			div <= "1110011110";
		else
			startup <= (others => '0');
			div <= (others => '0');
		end if;
  elsif(clk'event and clk = '1')then
    ddr_csn <= '0';
    if(startup = "11011" and tc = '1')then
      start_init <= '1';
      clke <= '1';
    else
      start_init <= '0';
    end if;
    ref <= tc and clke;
		if(init_done = '1')then
			startup <= (others => '1');
		elsif(tc = '1')then
			startup <= startup + 1;
		end if;
		if((not sim and div(9 downto 5) = "11101") or (sim and div(7) = '1'))then
			div <= (others => '0');
			tc <= '1';
		else
			div <= div + 1;
			tc <= '0';
		end if;
  end if;
end process;
end refresh_arch;
-----------------------------------------------------------------------------------
----------------------------------------------------------
library IEEE;
use IEEE.std_logic_1164.all;
library UNISIM;
use UNISIM.VComponents.all;

entity ddr_io is
    port (
        clk: in STD_LOGIC;
        clkn: in STD_LOGIC;
        pad: inout std_logic;
        t: in STD_LOGIC;
        d0: in STD_LOGIC;
        d1: in STD_LOGIC;
        iclk: in STD_LOGIC;
        iclkn: in STD_LOGIC;
        q0: out std_logic; -- rising edge       
        q1: out std_logic  -- falling edge    
    );
end ddr_io;

architecture ddr_io_arch of ddr_io is
signal din, oq, tq: std_logic;
attribute IOB : string;
attribute IOB of q0_reg,q1_reg,tq_reg: label is "TRUE";
begin
ib: ibuf port map(i => pad, o => din);
q0_reg: fd port map(d => din, c=> iclk, q => q0);
q1_reg: fd port map(d => din, c=> iclkn, q => q1);
oq_reg:  fddrrse
    port map(
          d0 => d0,
          d1 => d1,
          c0 => clk,
          c1 => clkn,
          ce => '1',
          r  => '0',
          s  => '0',
          q  => oq);
tq_reg:  fd
    port map(
          d => t,
          c => clk,
          q  => tq);
ob: obuft port map( i => oq, o => pad, t => tq);
end ddr_io_arch;
