----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    12:19:23 03/25/2007 
-- Design Name: 
-- Module Name:    sorterA - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
-- hit data from a DCON is delimted by a gap between event inputs of larger than 1ms to form a sorting group
-- maximum group size is 128 events.
-- If the group is bigger than 128 words, a group is splitted to subgroups of 128 events.
-- SorterA orders events within a subgroup or groups of up to 128 events according to their time stamp
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity sorterA is
    Port ( clk : in  STD_LOGIC;
           n_strobe : in  STD_LOGIC;
           reset : in  STD_LOGIC;
           skip : in  STD_LOGIC;
           one_ms : in  STD_LOGIC;
           DCON_data : in  STD_LOGIC_VECTOR (8 downto 0);
           DCON_we : in  STD_LOGIC;
           DCON_addr : in  STD_LOGIC_VECTOR (5 downto 0);
           hit_data : out  STD_LOGIC_VECTOR (8 downto 0);
           hit_addr : out  STD_LOGIC_VECTOR (1 downto 0);
           hit_end : out  STD_LOGIC; -- mark the end of a group, but not a subgroup
           hit_rqst : out  STD_LOGIC;
           hit_gnt : in  STD_LOGIC);
end sorterA;

architecture Behavioral of sorterA is
signal strobe : std_logic := '0';
signal rst_cycle : std_logic := '0';
signal fifo_do : std_logic_vector(9 downto 0) := (others => '0');
signal fifo_ra : std_logic_vector(10 downto 0) := (others => '0');
signal d_fifo_ra : std_logic_vector(4 downto 0) := (others => '0');
signal fifo_wa : std_logic_vector(10 downto 0) := (others => '0');
signal fifo_wap : std_logic_vector(4 downto 0) := (others => '0');
signal d_fifo_wa : std_logic_vector(4 downto 0) := (others => '0');
signal fifo_wa_addr : std_logic_vector(1 downto 0) := (others => '0');
signal ctrl_di : std_logic_vector(8 downto 0) := (others => '0');
signal ctrl_do : std_logic_vector(8 downto 0) := (others => '0');
signal ctrl_wa : std_logic_vector(10 downto 0) := (others => '0');
signal ctrl_ra : std_logic_vector(10 downto 0) := (others => '0');
signal we_ctrl : std_logic := '0';
signal buf_doa : std_logic_vector(9 downto 0) := (others => '0');
signal buf_dob : std_logic_vector(9 downto 0) := (others => '0');
signal buf_aa : std_logic_vector(12 downto 0) := (others => '0');
signal buf_ba : std_logic_vector(12 downto 0) := (others => '0');
signal d_buf_aa : std_logic_vector(1 downto 0) := (others => '0');
signal we_buf_aa : std_logic := '0';
signal we_buf : std_logic := '0';
begin
hit_data <= fifo_do(8 downto 0);
hit_end <= '0';
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
    if(strobe = '1')then
			if(DCON_we = '0')then
			end if;
		end if;
  end if;
end process;
fifo_wa(10 downto 9) <= DCON_addr(5 downto 4);
fifo_wa(3 downto 0) <= DCON_addr(3 downto 0);
i_fifo : RAMB16_S9_S9
   port map (
      DOB => fifo_do(7 downto 0),      -- Port B 8-bit Data Output
      DOPB => fifo_do(8 downto 8),      -- Port B 1-bit Data Output
      ADDRA => fifo_wa,  -- Port A 11-bit Address Input
      ADDRB => fifo_ra,  -- Port B 11-bit Address Input
      CLKA => clk,    -- Port A Clock
      CLKB => clk,    -- Port B Clock
      DIA => DCON_data(7 downto 0),      -- Port A 8-bit Data Input
      DIB => x"00",      -- Port B 8-bit Data Input
      DIPA => DCON_data(8 downto 8),    -- Port A 1-bit parity Input
      DIPB => "0",    -- Port-B 1-bit parity Input
      ENA => strobe,      -- Port A RAM Enable Input
      ENB => strobe,      -- PortB RAM Enable Input
      SSRA => '0',    -- Port A Synchronous Set/Reset Input
      SSRB => '0',    -- Port B Synchronous Set/Reset Input
      WEA => DCON_we,      -- Port A Write Enable Input
      WEB => '0'       -- Port B Write Enable Input
   );
fifo_do(9) <= '0';
g_fifo_wa: for i in 0 to 4 generate
  i_fifo_wa : RAM16X1D
   port map (
      SPO => fifo_wa(i+4),     -- Read-only 1-bit data output for DPRA
      DPO => fifo_wap(i),     -- Read-only 1-bit data output for DPRA
      A0 => fifo_wa_addr(0),       -- R/W address[0] input bit
      A1 => fifo_wa_addr(1),       -- R/W address[1] input bit
      A2 => '0',       -- R/W address[2] input bit
      A3 => '0',       -- R/W ddress[3] input bit
      D => d_fifo_wa(i),         -- Write 1-bit data input
      DPRA0 => fifo_ra(9), -- Read-only address[0] input bit
      DPRA1 => fifo_ra(10), -- Read-only address[1] input bit
      DPRA2 => '0', -- Read-only address[2] input bit
      DPRA3 => '0', -- Read-only address[3] input bit
      WCLK => clk,   -- Write clock input
      WE => strobe        -- Write enable input
   );
end generate;
g_fifo_ra: for i in 0 to 4 generate
  i_fifo_ra : RAM16X1S
   port map (
      O => fifo_ra(i+4),     -- RAM data output
      A0 => fifo_ra(9),       -- RAM address[0] input bit
      A1 => fifo_ra(10),       -- RAM address[1] input bit
      A2 => '0',       -- RAM address[2] input bit
      A3 => '0',       -- RAM ddress[3] input bit
      D => d_fifo_ra(i),         -- RAM 1-bit data input
      WCLK => clk,   -- Write clock input
      WE => strobe        -- Write enable input
   );
end generate;
d_fifo_wa <= (others => '0') when rst_cycle = '1' else
						fifo_wa(8 downto 4) + 1 when DCON_we = '1' and DCON_addr(3 downto 0) = x"f" else
						fifo_wa(8 downto 4);
d_fifo_ra <= (others => '0') when rst_cycle = '1' else
						fifo_ra(8 downto 4) + 1 when fifo_ra(3 downto 0) = x"f" else
						fifo_ra(8 downto 4);
process(clk)
begin
  if(clk'event and clk = '1')then
    if(strobe = '1')then
			if(reset = '1')then
				rst_cycle <= '1';
			elsif(fifo_wa_addr = "11")then
				rst_cycle <= '0';
			end if;
			if(reset = '1')then
				fifo_wa_addr <= "00";
			elsif(rst_cycle = '1')then
				fifo_wa_addr <= fifo_wa_addr + 1;
			else
				fifo_wa_addr <= DCON_addr(5 downto 4);
			end if;
			if(rst_cycle = '1')then
				fifo_ra(3 downto 0) <= x"0";
			elsif(hit_gnt = '1')then
				fifo_ra(3 downto 0) <= fifo_ra(3 downto 0) + 1;
			end if;
			if(rst_cycle = '1' or hit_gnt = '1')then
				hit_rqst <= '0';
			elsif(fifo_ra(8 downto 4) /= fifo_wap)then
				hit_rqst <= '1';
			end if;
			if(fifo_ra(8 downto 4) = fifo_wap)then
				fifo_ra(10 downto 9) <= fifo_ra(10 downto 9) + 1;
--		  fifo_ra(3 downto 0) <= x"0";
--		  we_buf <= '0';
			else
				hit_addr <= fifo_ra(10 downto 9);
--		  fifo_ra(3 downto 0) <= fifo_ra(3 downto 0) + 1;
--		  we_buf <= '1';
			end if;
--		buf_aa(12 downto 11) <= fifo_ra(10 downto 9);
--		buf_aa(8 downto 0) <= fifo_ra(8 downto 0);
--	   if(rst_cycle = '1' or (fifo_ra(8 downto 0) = "111111111" and we_buf = '1'))then
--		  we_buf_aa <= '1';
--		else
--		  we_buf_aa <= '0';
--		end if;
--		if(rst_cycle = '1')then
--		  d_buf_aa <= (others => '0');
--		else
--		  d_buf_aa <= buf_aa(10 downto 9) + 1;
--		end if;
--		if(hit_gnt = '1')then
--		  buf_ba(3 downto 0) <= buf_ba(3 downto 0) + 1;
--		end if;
		end if;
  end if;
end process;
--g_buf_aa: for i in 0 to 1 generate
--  i_buf_aa : RAM16X1S
--   port map (
--      O => buf_aa(i+9),     -- RAM data output
--      A0 => buf_aa(11),       -- RAM address[0] input bit
--      A1 => buf_aa(12),       -- RAM address[1] input bit
--      A2 => '0',       -- RAM address[2] input bit
--      A3 => '0',       -- RAM ddress[3] input bit
--      D => d_buf_aa(i),         -- RAM 1-bit data input
--      WCLK => clk,   -- Write clock input
--      WE => we_buf_aa        -- Write enable input
--   );
--end generate;
--i_ctrl : RAMB16_S9_S9
--   port map (
--      DOB => ctrl_do(7 downto 0),      -- Port B 8-bit Data Output
--      DOPB => ctrl_do(8 downto 8),      -- Port B 1-bit Data Output
--      ADDRA => ctrl_wa,  -- Port A 11-bit Address Input
--      ADDRB => ctrl_ra,  -- Port B 11-bit Address Input
--      CLKA => clk,    -- Port A Clock
--      CLKB => clk,    -- Port B Clock
--      DIA => ctrl_di(7 downto 0),      -- Port A 8-bit Data Input
--      DIB => x"00",      -- Port B 8-bit Data Input
--      DIPA => ctrl_di(8 downto 8),    -- Port A 1-bit parity Input
--      DIPB => "0",    -- Port-B 1-bit parity Input
--      ENA => strobe,      -- Port A RAM Enable Input
--      ENB => strobe,      -- PortB RAM Enable Input
--      SSRA => '0',    -- Port A Synchronous Set/Reset Input
--      SSRB => '0',    -- Port B Synchronous Set/Reset Input
--      WEA => we_ctrl,      -- Port A Write Enable Input
--      WEB => '0'       -- Port B Write Enable Input
--   );
--g_buf : for i in 0 to 4 generate
--  i_buf : RAMB16_S2_S2
--    port map (
--      DOA => buf_doa(2*i+1 downto 2*i),      -- Port A 2-bit Data Output
--      DOB => buf_dob(2*i+1 downto 2*i),      -- Port B 2-bit Data Output
--      ADDRA => buf_aa,  -- Port A 13-bit Address Input
--      ADDRB => buf_ba,  -- Port B 13-bit Address Input
--      CLKA => clk,    -- Port A Clock
--      CLKB => clk,    -- Port B Clock
--      DIA => fifo_do(2*i+1 downto 2*i),      -- Port A 2-bit Data Input
--      DIB => "00",      -- Port B 2-bit Data Input
--      ENA => strobe,      -- Port A RAM Enable Input
--      ENB => strobe,      -- PortB RAM Enable Input
--      SSRA => '0',    -- Port A Synchronous Set/Reset Input
--      SSRB => '0',    -- Port B Synchronous Set/Reset Input
--      WEA => we_buf,      -- Port A Write Enable Input
--      WEB => '0'       -- Port B Write Enable Input
--   );
--end generate;
end Behavioral;

