----------------------------------------------------------------------------------
-- Company: 
-- Engineer: 
-- 
-- Create Date:    12:19:23 03/25/2007 
-- Design Name: 
-- Module Name:    sorterA - Behavioral 
-- Project Name: 
-- Target Devices: 
-- Tool versions: 
-- Description: 
--
-- Dependencies: 
--
-- Revision: 
-- Revision 0.01 - File Created
-- Additional Comments: 
-- SorterC sorts events from different channels.
-- measures must taken in order not to confuse events because of time stamp wrap around
----------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

---- Uncomment the following library declaration if instantiating
---- any Xilinx primitives in this code.
library UNISIM;
use UNISIM.VComponents.all;

entity sorterC is
    Port ( clk : in  STD_LOGIC;
           n_strobe : in  STD_LOGIC;
           reset : in  STD_LOGIC;
           circular : in  STD_LOGIC;
           buf_full : in  STD_LOGIC;
           one_ms : in  STD_LOGIC;
           ch_data : in  STD_LOGIC_VECTOR (31 downto 0);
           ch_addr : in  STD_LOGIC_VECTOR (5 downto 0);
           ch_dav : in  STD_LOGIC_VECTOR (11 downto 0); -- flag for more data of the channel.
           ch_we : in  STD_LOGIC; -- each time writes a burst of 4*16 bytes
           ch_gnt : in  STD_LOGIC; -- request queued
           ch_rqst : out  STD_LOGIC; -- request channel data
           ch_num : out  STD_LOGIC_VECTOR (3 downto 0);  -- requested channel number
           evt_data : out  STD_LOGIC_VECTOR (7 downto 0); -- sorted data out
			  evt_end : out STD_LOGIC; -- end of event mark in non-circular mode
           evt_we : out  STD_LOGIC); -- write enable
end sorterC;

architecture Behavioral of sorterC is
signal strobe : std_logic := '0';
signal rst_cycle : std_logic := '0';
signal fifo_do : std_logic_vector(7 downto 0) := (others => '0');
signal fifo_wa : std_logic_vector(9 downto 0) := (others => '0');
signal fifo_ra : std_logic_vector(11 downto 0) := (others => '0');
signal d_fifo_wa : std_logic_vector(1 downto 0) := (others => '0');
signal fifo_wap : std_logic_vector(1 downto 0) := (others => '0');
signal ch_num_i : std_logic_vector(3 downto 0) := (others => '0');
signal ch_queued : std_logic := '0';
signal queue_wa : std_logic_vector(3 downto 0) := (others => '0');
signal d_queue : std_logic := '0';
signal we_fifo : std_logic := '0';
signal evt_we_i : std_logic := '0';
signal ch_rqst_i : std_logic := '0';
signal d_fifo_ra : std_logic_vector(1 downto 0) := (others => '0');
signal ctrl_di : std_logic_vector(8 downto 0) := (others => '0');
signal ctrl_do : std_logic_vector(8 downto 0) := (others => '0');
signal ctrl_wa : std_logic_vector(10 downto 0) := (others => '0');
signal ctrl_ra : std_logic_vector(10 downto 0) := (others => '0');
signal we_ctrl : std_logic := '0';
signal buf_doa : std_logic_vector(7 downto 0) := (others => '0');
signal buf_dob : std_logic_vector(7 downto 0) := (others => '0');
signal buf_aa : std_logic_vector(10 downto 0) := (others => '0');
signal buf_ab : std_logic_vector(10 downto 0) := (others => '0');
signal we_buf : std_logic := '0';
begin
ch_rqst <= ch_rqst_i;
ch_num <= ch_num_i;
evt_we <= evt_we_i;
evt_data <= fifo_do;
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
    if(strobe = '1')then
	   if(circular = '0' and ch_dav = x"000")then -- should add last_hit condition
		  evt_end <= '1';
		else
		  evt_end <= '0';
		end if;
		d_queue <= ch_gnt and not rst_cycle;
		if(rst_cycle = '1')then
		  queue_wa <= fifo_wa(7 downto 4);
		elsif(ch_gnt = '1')then
		  queue_wa <= ch_num_i;
		elsif(ch_we = '1')then
		  queue_wa <= fifo_wa(7 downto 4);
		else
		  queue_wa <= x"f";
		end if;
	 end if;
  end if;
end process;
i_ch_queued : RAM16X1D
   port map (
      DPO => ch_queued,     -- Read-only 1-bit data output for DPRA
      A0 => queue_wa(0),       -- R/W address[0] input bit
      A1 => queue_wa(1),       -- R/W address[1] input bit
      A2 => queue_wa(2),       -- R/W address[2] input bit
      A3 => queue_wa(3),       -- R/W ddress[3] input bit
      D => d_queue,         -- Write 1-bit data input
      DPRA0 => fifo_ra(6), -- Read-only address[0] input bit
      DPRA1 => fifo_ra(7), -- Read-only address[1] input bit
      DPRA2 => fifo_ra(8), -- Read-only address[2] input bit
      DPRA3 => fifo_ra(9), -- Read-only address[3] input bit
      WCLK => clk,   -- Write clock input
      WE => strobe       -- Write enable input
   );
process(clk)
variable fifo_wc : std_logic_vector(3 downto 0);
begin
  fifo_wc := fifo_ra(11 downto 10) & fifo_wap;
  if(clk'event and clk = '1')then
    if(strobe = '1')then
	   if(reset = '1')then
		  rst_cycle <= '1';
		elsif(fifo_wa(7 downto 4) = x"b")then
		  rst_cycle <= '0';
		end if;
		if(rst_cycle = '1')then
		  ch_rqst_i <= '0';
		elsif(ch_rqst_i = '0' and ch_queued = '0' and ch_dav(conv_integer(fifo_ra(9 downto 6))) = '1')then
		  case fifo_wc is
		    when x"0" | x"1" | x"5" | x"6" | x"a" | x"b" | x"c" | x"f" => ch_rqst_i <= '1';
																								ch_num_i <= fifo_ra(9 downto 6);
		    when others => ch_rqst_i <= '0';
		  end case;
		elsif(ch_gnt = '1')then
		  ch_rqst_i <= '0';
		end if;
		if(reset = '1')then
		  fifo_wa(7 downto 4) <= x"0";
		elsif(rst_cycle = '1')then
		  fifo_wa(7 downto 4) <= fifo_wa(7 downto 4) + 1;
		else 
		  fifo_wa(7 downto 4) <= ch_addr(5 downto 2);
		end if;
		if(reset = '1' or fifo_ra(9 downto 6) = x"b")then
		  fifo_ra(9 downto 6) <= x"0";
		elsif(rst_cycle = '1' or fifo_ra(11 downto 10) = fifo_wap)then
		  fifo_ra(9 downto 6) <= fifo_ra(9 downto 6) + 1;
		end if;
		if(rst_cycle = '1' or fifo_ra(3 downto 0) = x"f")then
		  evt_we_i <= '0';
		elsif(fifo_ra(11 downto 10) /= fifo_wap and buf_full = '0')then
		  evt_we_i <= '1';
		end if;
		if(evt_we_i = '0')then
		  fifo_ra(3 downto 0) <= x"0";
		else
		  fifo_ra(3 downto 0) <= fifo_ra(3 downto 0) + 1;
		end if;
      fifo_wa(1 downto 0) <= ch_addr(1 downto 0);
		if(ch_we = '0')then
		  fifo_wa(3 downto 2) <= "00";
		elsif(fifo_wa(1 downto 0) = "11")then
		  fifo_wa(3 downto 2) <= fifo_wa(3 downto 2) + 1;
		end if;
		we_fifo <= ch_we;
	 end if;
  end if;
end process;
d_fifo_wa <= "00" when rst_cycle = '1' else fifo_wa(9 downto 8) + 1 when fifo_wa(3 downto 0) = x"f" else fifo_wa(9 downto 8);
d_fifo_ra <= "00" when rst_cycle = '1' else fifo_ra(11 downto 10) + 1 when fifo_ra(3 downto 0) = x"f" else fifo_ra(11 downto 10);
g_fifo_wa : for i in 0 to 1 generate
  i_fifo_wa : RAM16X1D
   port map (
      SPO => fifo_wa(8+i),     -- Read-only 1-bit data output for DPRA
      DPO => fifo_wap(i),     -- Read-only 1-bit data output for DPRA
      A0 => fifo_wa(4),       -- R/W address[0] input bit
      A1 => fifo_wa(5),       -- R/W address[1] input bit
      A2 => fifo_wa(6),       -- R/W address[2] input bit
      A3 => fifo_wa(7),       -- R/W ddress[3] input bit
      D => d_fifo_wa(i),         -- Write 1-bit data input
      DPRA0 => fifo_ra(6), -- Read-only address[0] input bit
      DPRA1 => fifo_ra(7), -- Read-only address[1] input bit
      DPRA2 => fifo_ra(8), -- Read-only address[2] input bit
      DPRA3 => fifo_ra(9), -- Read-only address[3] input bit
      WCLK => clk,   -- Write clock input
      WE => strobe        -- Write enable input
   );
end generate;
g_fifo_ra : for i in 0 to 1 generate
  i_fifo_ra : RAM16X1S
   port map (
      O => fifo_ra(10 + i),     -- RAM data output
      A0 => fifo_ra(6),       -- RAM address[0] input bit
      A1 => fifo_ra(7),       -- RAM address[1] input bit
      A2 => fifo_ra(8),       -- RAM address[2] input bit
      A3 => fifo_ra(9),       -- RAM ddress[3] input bit
      D => d_fifo_ra(i),         -- RAM 1-bit data input
      WCLK => clk,   -- Write clock input
      WE => strobe        -- Write enable input
   );
end generate;
i_fifo : RAMB16_S9_S36
   port map (
      DOA => fifo_do,      -- Port A 8-bit Data Output
      ADDRA => fifo_ra(10 downto 0),  -- Port A 11-bit Address Input
      ADDRB => fifo_wa(8 downto 0),  -- Port B 9-bit Address Input
      CLKA => clk,    -- Port A Clock
      CLKB => clk,    -- Port B Clock
      DIA => x"00",      -- Port A 8-bit Data Input
      DIB => ch_data,      -- Port B 32-bit Data Input
      DIPA => "0",    -- Port A 1-bit parity Input
      DIPB => x"0",    -- Port-B 4-bit parity Input
      ENA => strobe,      -- Port A RAM Enable Input
      ENB => strobe,      -- PortB RAM Enable Input
      SSRA => '0',    -- Port A Synchronous Set/Reset Input
      SSRB => '0',    -- Port B Synchronous Set/Reset Input
      WEA => '0',      -- Port A Write Enable Input
      WEB => ch_we       -- Port B Write Enable Input
   );
--i_buf : RAMB16_S9_S9
--   port map (
--      DOA => buf_doa,      -- Port A 8-bit Data Output
--      DOB => buf_dob,      -- Port B 8-bit Data Output
--      ADDRA => buf_aa,  -- Port A 11-bit Address Input
--      ADDRB => buf_ab,  -- Port B 11-bit Address Input
--      CLKA => clk,    -- Port A Clock
--      CLKB => clk,    -- Port B Clock
--      DIA => fifo_do,      -- Port A 8-bit Data Input
--      DIB => x"00",      -- Port B 8-bit Data Input
--      DIPA => "0",    -- Port A 1-bit parity Input
--      DIPB => "0",    -- Port-B 1-bit parity Input
--      ENA => strobe,      -- Port A RAM Enable Input
--      ENB => strobe,      -- PortB RAM Enable Input
--      SSRA => '0',    -- Port A Synchronous Set/Reset Input
--      SSRB => '0',    -- Port B Synchronous Set/Reset Input
--      WEA => we_buf,      -- Port A Write Enable Input
--      WEB => '0'       -- Port B Write Enable Input
--   );
end Behavioral;

