-- if we_byte are not all one, the write cycle takes four cycles to
-- allow read back the old data for those unchanged bytes
library IEEE,mylib;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
use mylib.DCOL_pack.all;
library UNISIM;
use UNISIM.VComponents.all;

entity vme_if is
    generic (swap_byte : boolean := false; swap_lw : boolean := false);
    Port ( vme_pads : inout vme_pad;
           SW_BAR : in std_logic_vector(4 downto 0);
           SW_use : in std_logic;
           clk : in std_logic;
           n_strobe : in std_logic;
           vme_BLT : out std_logic;
           ram_dav : in std_logic;
           rfifo_afull : out std_logic;
           rd_ram : out std_logic;
           sel_mod : out std_logic;
           sel_reg : out std_logic;
           r_data : in std_logic_vector(31 downto 0);
           w_data : out std_logic_vector(31 downto 0);
           we_byte : out std_logic_vector(3 downto 0);
           sys_reset : out std_logic; -- vme system reset
           addr : out std_logic_vector(22 downto 0)
			  );
end vme_if;

architecture my_arch of vme_if is
signal strobe : std_logic := '0';
signal last_word : std_logic := '0';
signal MBLT : std_logic := '0';
signal MBLT_data : std_logic := '0';
signal BLT : std_logic := '0';
signal vme_reset : std_logic := '0';
signal vme_reset_sync : std_logic := '0';
signal asn : std_logic := '1';
signal asn_latch : std_logic := '1';
signal viackn : std_logic := '1';
signal va : std_logic_vector(31 downto 0) := (others => '0');
signal vam : std_logic_vector(5 downto 0) := (others => '0');
signal BAR : std_logic_vector(6 downto 0) := (others => '0');
signal dsn_q : std_logic_vector(1 downto 0) := (others => '0');
signal dsn0_sync : std_logic_vector(1 downto 0) := (others => '0');
signal dsn1_sync : std_logic := '0';
signal sel_mod_i : std_logic := '0';
signal sel_reg_i : std_logic := '0';
signal sel_mod_dl : std_logic := '0';
signal w_ack : std_logic := '0';
signal we_byte_i : std_logic_vector(3 downto 0) := (others => '0');
signal en_write : std_logic := '0';
signal rd_DONE : std_logic := '0';
signal rd_rqst : std_logic_vector(1 downto 0) := (others => '0');
signal rd_ram_i : std_logic := '0';
signal rst_rfifo : std_logic := '0';
signal we_rfifo : std_logic := '0';
signal rd_rfifo : std_logic := '0';
signal rfifo_dav : std_logic := '0';
signal rfifo_afull_i : std_logic := '0';
signal rfifo_rp : std_logic_vector(3 downto 0) := (others => '0');
signal r_cycle : std_logic := '0';
signal D32 : std_logic := '0';
signal dtack : std_logic := '0';
signal dtack_dl : std_logic_vector(1 downto 0) := (others => '0');
signal t_dtack : std_logic := '1';
signal vap : std_logic_vector(24 downto 1) := (others => '0');
signal vme_data : std_logic_vector(63 downto 0) := (others => '0');
signal rfifo_dout : std_logic_vector(31 downto 0) := (others => '0');
signal rfifo_din : std_logic_vector(31 downto 0) := (others => '0');
signal w_data_i : std_logic_vector(31 downto 0) := (others => '0');
attribute IOB: string;
attribute IOB of dsn_q : signal is "true";
attribute LOC: string;
--attribute LOC of i_asn : label is "";
--attribute iob of asn_latch : signal is "true";
--attribute init : string;
--attribute init of asn_latch : signal is "s";
begin
-- endianess swapping
w_data <= w_data_i when swap_byte = false else w_data_i(7 downto 0) & w_data_i(15 downto 8) & w_data_i(23 downto 16) & w_data_i(31 downto 24);
rfifo_din <= r_data when swap_byte = false else r_data(7 downto 0) & r_data(15 downto 8) & r_data(23 downto 16) & r_data(31 downto 24);
we_byte <= we_byte_i when swap_byte = false else we_byte_i(0) & we_byte_i(1) & we_byte_i(2) & we_byte_i(3);
-- endianess swapping
sel_mod <= sel_mod_i;
i_asn: bufg port map(i => vme_pads.asn, o => asn);
rfifo_afull <= rfifo_afull_i;
rd_ram <= rd_ram_i;
vme_pads.irq1n <= '1';
vme_pads.irq2n <= '1';
vme_pads.irq4n <= '1';
vme_pads.irq7n <= '1';
vme_pads.iackoutn <= vme_pads.iackinn;
vme_pads.oen <= '0';
vme_pads.sysfailn <= '1';
vme_pads.berrn <= '1';
addr <= vap(24 downto 2);
sel_reg <= sel_reg_i;
MBLT <= '1' when vam(1 downto 0) = "00" else '0';
BLT <= '1' when vam(1) = vam(0) else '0'; -- include MBLT
-- synchronize DSn signals
-- synchronize vme_reset
process(clk)
begin
  if(clk'event and clk = '1')then
    strobe <= n_strobe;
    dsn_q <= vme_pads.dsn;
    dsn0_sync(0) <= dsn_q(0);
    dsn1_sync <= dsn_q(1);
    if(strobe = '1')then
      dsn0_sync(1) <= dsn0_sync(0);
      vme_reset_sync <= vme_pads.resetn;
      vme_reset <= not vme_reset_sync;
      vme_BLT <= BLT and sel_mod_i;
		end if;
  end if;
end process;
-- vme system reset signal
sys_reset <= vme_reset;
-- latching AM and Address lines
process(asn,vme_pads.a,vme_pads.am,vme_pads.iackn)
begin
  if(asn = '1')then
    va <= vme_pads.a;
		vam <= vme_pads.am;
		viackn <= vme_pads.iackn;
  end if;
end process;
BAR(6 downto 2) <= SW_BAR when SW_use = '1' or vme_pads.gan = "11111" else not vme_pads.gan;
BAR(1 downto 0) <= "00";
asn_latch <= vme_pads.asn;
process(clk,asn_latch)
begin
  if(asn_latch = '1')then
    sel_mod_i <= '0';
    sel_mod_dl <= '0';
  elsif(clk'event and clk = '1')then
    if(strobe = '1')then
      if(viackn = '1' and vam(5 downto 3) = "001" and BAR = va(31 downto 25))then
				sel_mod_i <= '1';
			end if;
			sel_mod_dl <= sel_mod_i;
		end if;
  end if;
end process;
-- dsn synchronisation and ds rising edge detection
process(clk)
begin
  if(clk'event and clk = '1')then
    if(sel_mod_i = '0')then
			last_word <= not MBLT;
		elsif((strobe = '1' and we_byte_i(0) = '1') or rd_rfifo = '1')then
			last_word <= not last_word or not MBLT;
		end if;	 
    if(sel_mod_i = '0')then
			MBLT_data <= '0';
		elsif(dtack_dl = "10" and MBLT = '1')then
			MBLT_data <= '1';
		end if;	 
    if(strobe = '1')then
      if(sel_mod_i = '0')then
				sel_reg_i <= '0';
			elsif(va(24 downto 6) = "0000000000000000000")then
				sel_reg_i <= '1';
			end if;
      if(sel_mod_i = '0')then
				rd_ram_i <= '0';
				vap <= (others => '0');
			else
				rd_ram_i <= vme_pads.writen and (va(24) or va(23) or va(14));
        if(sel_mod_dl = '0')then
					vap(24 downto 1) <= va(24 downto 1); 
				else
					if(rd_rqst(0) = '1' or we_byte_i(3) = '1')then
						vap(10 downto 2) <= vap(10 downto 2) + 1; 
					end if;
					if(va(0) = '1' and dsn0_sync = "01")then
						vap(1) <= not vap(1); 
					end if;
				end if;
			end if;
		end if;
    if(dtack = '1')then
			t_dtack <= '0';
		elsif(dtack_dl = "00")then
			t_dtack <= '1';
		end if;
		dtack_dl <= dtack_dl(0) & dtack;
  end if;
end process;
vme_pads.t_dtack <= t_dtack;
process(clk,dsn_q)
begin
  if(dsn_q = "11")then
    vme_pads.dtackn <= '1';
    dtack <= '0';
  elsif(clk'event and clk = '1')then
    vme_pads.dtackn <= not dtack;
		if((((last_word = '1' and rfifo_dav = '1') or (MBLT = '1' and MBLT_data = '0' and sel_mod_i = '1')) and 
	    (dsn0_sync(0) = '0' or dsn1_sync = '0')) or (last_word = '1' and n_strobe = '0' and en_write = '1'))then
			dtack <= '1';
		end if;
  end if;
end process; 
-- rescinding DTACKn driven high for two clock cycles(16ns) before going 3-stated;
process(clk)
begin
  if(clk'event and clk = '1')then
		if(sel_mod_i = '0')then
			rd_DONE <= '0';
		elsif(BLT = '0' and rd_rqst(0) = '1')then
			rd_DONE <= '1';
		end if;
		if(sel_mod_i = '0' or rfifo_rp = x"5")then
			rfifo_afull_i <= '0';
		elsif(rfifo_rp = x"6")then
			rfifo_afull_i <= '1';
		end if;
		if(vme_pads.writen = '0' or sel_mod_dl = '0')then
			rd_rqst <= "00";
		elsif(strobe = '1')then
			rd_rqst(0) <= (not rfifo_afull_i and not rd_done and not rd_ram_i) or ram_dav;
      rd_rqst(1) <= rd_rqst(0); 
		end if;
		if(sel_mod_i = '0')then
			D32 <= '0';
		elsif(dsn0_sync(0) = '0' and dsn1_sync = '0' and va(1 downto 0) = "00")then
			D32 <= '1';
		end if;
		if(sel_mod_i = '0' or vme_pads.writen = '1')then
			en_write <= '0';
		elsif(strobe = '1')then
			if(w_ack = '1')then
				en_write <= '0';
			elsif((dsn0_sync(0) = '0' or dsn1_sync = '0') and (MBLT = '0' or MBLT_data = '1'))then
				en_write <= '1';
			end if;
		end if;
		if(dsn0_sync(0) = '1' and dsn1_sync = '1')then
			w_ack <= '0';
		elsif(last_word = '1' and en_write = '1')then
			w_ack <= '1';
		end if;
-- byte lane select !!! NOTICE byte(0) is the MSB !!!
-- for 16-bit or 8-bit registers, select appropriate we_byte signal
-- according to the address 
		if(strobe = '1')then
			if(en_write = '0' or vap(24 downto 15) /= "0000000000" or we_byte_i(0) = '1')then -- circular buffer not writable
				we_byte_i <= x"0";
			elsif(D32 = '1')then
				we_byte_i <= x"f";
			else
				we_byte_i(0) <= not dsn_q(1) and not vap(1);
				we_byte_i(1) <= (not dsn_q(0) and not vap(1)) or not va(0);
				we_byte_i(2) <= (not dsn_q(1) and vap(1)) or not va(0);
				we_byte_i(3) <= not dsn_q(0) and (vap(1) xnor va(0));
			end if;
		end if;
  end if;
end process;
rst_rfifo <= not sel_mod_i;
i_rfifo: srl_fifo PORT MAP(
		rst => rst_rfifo,
		din => rfifo_din,
		we => we_rfifo,
		clk => clk,
		re => rd_rfifo,
		dav => rfifo_dav,
		dout => rfifo_dout,
		rp => rfifo_rp
	);
-- D16(D8) data swapping
process(clk)
begin
  if(clk'event and clk = '1')then
    if(sel_reg_i = '1' or rd_ram_i = '1')then
			we_rfifo <= rd_rqst(0) and n_strobe;
		else
			we_rfifo <= rd_rqst(1) and n_strobe;
		end if;	 
    if(rfifo_dav = '1' and rd_rfifo = '0' and ((last_word = '0' and MBLT_data = '1') or 
	    (dsn0_sync = "01" and (D32 = '1' or vap(1) = '1'))))then
			rd_rfifo <= '1';
		else
			rd_rfifo <= '0';
		end if;
		if(swap_lw = true or (last_word = '0' and rd_rfifo = '1'))then
      vme_data(63 downto 32) <= rfifo_dout;
		end if;
		if(swap_lw = true and last_word = '0' and rd_rfifo = '1')then
      vme_data(31 downto 0) <= rfifo_dout;
    elsif(dtack = '0')then
      vme_data(31 downto 16) <= rfifo_dout(31 downto 16);
			if(vap(1) = '0' and va(0) = '1')then
				vme_data(15 downto 0) <= rfifo_dout(31 downto 16);
			else
				vme_data(15 downto 0) <= rfifo_dout(15 downto 0);
			end if;
		end if;
		if(sel_mod_i = '0' or r_cycle = '1')then
			w_data_i <= (others => '0');
		elsif(strobe = '1')then
			if(last_word = '0')then
				if(swap_lw = false)then
					w_data_i <= vme_pads.a;
				else
					w_data_i <= vme_pads.d;
				end if;
			elsif(MBLT = '1' and swap_lw = true)then
				w_data_i <= vme_pads.a;
			elsif(va(0) = '0')then
				w_data_i <= vme_pads.d;
			else
				w_data_i <= vme_pads.d(15 downto 0) & vme_pads.d(15 downto 0);
			end if;
		end if;
		if(dtack = '0')then
			r_cycle <= vme_pads.writen and sel_mod_i and (not MBLT or MBLT_data);
		end if;
  end if;
end process;
vme_pads.d_dir <= r_cycle;
vme_pads.a_dir <= MBLT_data and vme_pads.writen and sel_mod_i; 
vme_pads.a <= (others => 'Z') when MBLT_data  = '0' or vme_pads.writen = '0' or dsn_q(0) = '1' else vme_data(63 downto 32); 
vme_pads.d <= (others => 'Z') when r_cycle = '0' or dsn_q = "11" else vme_data(31 downto 0); 
end my_arch;
